from os import PathLike
from typing import Any, BinaryIO, Union

from ..serializer import MaybeHint, REGISTRY, Hint, WrongSerializer
from .helpers import ExtensionMatch, SourceAgnostic


class ImageIO(ExtensionMatch, SourceAgnostic):
    extensions = '.png', '.jpg', '.tif', '.bmp',

    def _match_value(self, value):
        return isinstance(value, np.ndarray)

    def load(self, source: Union[PathLike, BinaryIO], hint: MaybeHint, params) -> Any:
        if hint is not None:
            hint = '.' + hint.split('.')[-1]
        return imread(source, extension=hint)

    def save(self, value: Any, destination: Union[PathLike, BinaryIO], hint: MaybeHint, params) -> Hint:
        if hint is None:
            raise WrongSerializer

        hint = '.' + hint.split('.')[-1]
        imwrite(destination, value, extension=hint)
        return hint


try:
    try:
        from imageio.v3 import imread, imwrite
    except ImportError:
        # py3.6
        from imageio import imread as _imread, imwrite as _imwrite


        def imread(x, extension):
            return _imread(x, format=extension)


        def imwrite(x, y, extension):
            return _imwrite(x, y, format=extension)

    import numpy as np

    REGISTRY.append(ImageIO())

except ImportError:
    pass



from os import PathLike
from typing import BinaryIO, Any, Union

from ..serializer import MaybeHint, REGISTRY, Hint
from .helpers import ExtensionMatch, SourceAgnostic
from .packaged import Gzip


class DICOM(ExtensionMatch, SourceAgnostic):
    extensions = '.dcm',

    def _match_value(self, value):
        return isinstance(value, pydicom.Dataset)

    def load(self, source: Union[PathLike, BinaryIO], hint: MaybeHint, params) -> Any:
        return pydicom.dcmread(source)

    def save(self, value: Any, destination: Union[PathLike, BinaryIO], hint: MaybeHint, params) -> Hint:
        pydicom.dcmwrite(destination, value)
        return '.dcm'


try:
    import pydicom

    REGISTRY.append(DICOM())
    REGISTRY.append(Gzip(DICOM()))

except ImportError:
    pass



from abc import ABC, abstractmethod
from enum import Enum
from os import PathLike
from pathlib import Path
from typing import Any, Union, BinaryIO

Hint = str
MaybeHint = Union[str, None]
MaybePath = Union[Path, None]
REGISTRY = []


class MatchHint(Enum):
    Accept, NotSure, Reject = 0, 1, 2


class Serializer(ABC):
    # save
    @abstractmethod
    def match_save_buffer(self, value: Any, hint: MaybeHint, params: dict) -> MatchHint:
        pass

    @abstractmethod
    def save_buffer(self, value: Any, destination: BinaryIO, hint: MaybeHint, params: dict) -> Hint:
        pass

    @abstractmethod
    def match_save_path(self, value: Any, destination: PathLike, hint: Hint, params: dict) -> MatchHint:
        pass

    @abstractmethod
    def save_path(self, value: Any, destination: PathLike, hint: Hint, params: dict) -> Hint:
        pass

    # load
    @abstractmethod
    def match_load_buffer(self, hint: MaybeHint, allow_lazy: bool, params: dict) -> MatchHint:
        pass

    @abstractmethod
    def load_buffer(self, source: BinaryIO, hint: MaybeHint, allow_lazy: bool, params: dict) -> Any:
        pass

    @abstractmethod
    def match_load_path(self, source: PathLike, hint: Hint, params: dict) -> MatchHint:
        pass

    @abstractmethod
    def load_path(self, source: PathLike, hint: Hint, params: dict) -> Any:
        pass


class WrongSerializer(Exception):
    pass


class RequireLazy(Exception):
    pass



from os import PathLike
from typing import BinaryIO, Any, Union

from ..serializer import MaybeHint, REGISTRY, Hint
from .helpers import ExtensionMatch, SourceAgnostic
from .packaged import Gzip


class CSV(ExtensionMatch, SourceAgnostic):
    extensions = '.csv',

    def _match_value(self, value):
        return isinstance(value, (DataFrame, Series))

    def _match_save_params(self, params: dict):
        return set(params) <= {'index'}

    def load(self, source: Union[PathLike, BinaryIO], hint: MaybeHint, params) -> Any:
        return pd.read_csv(source)

    def save(self, value: Any, destination: Union[PathLike, BinaryIO], hint: MaybeHint, params) -> Hint:
        value.to_csv(destination, **params)
        return '.csv'


try:
    import pandas as pd
    from pandas import DataFrame, Series

    REGISTRY.append(CSV())
    REGISTRY.append(Gzip(CSV()))

except ImportError:
    pass



from os import PathLike
from typing import Any, BinaryIO, Union

from . import ExtensionMatch
from .helpers import SourceAgnostic
from ..serializer import MaybeHint, REGISTRY, Hint
from .packaged import Gzip


class Numpy(SourceAgnostic, ExtensionMatch):
    extensions = '.npy',

    def _match_value(self, value):
        return isinstance(value, (np.ndarray, np.generic))

    def load(self, source: Union[PathLike, BinaryIO], hint: MaybeHint, params) -> Any:
        return np.load(source, allow_pickle=False)

    def save(self, value: Any, destination: Union[PathLike, BinaryIO], hint: MaybeHint, params) -> Hint:
        np.save(destination, value, allow_pickle=False)
        return '.npy'


try:
    import numpy as np

    REGISTRY.append(Numpy())
    REGISTRY.append(Gzip(Numpy()))

except ImportError:
    pass



import json
import os
import pickle
from gzip import GzipFile
from os import PathLike
from typing import Any, Union, BinaryIO

from .serializer import MaybeHint, WrongSerializer, REGISTRY, Hint
from .serializers.choice import Choice

__all__ = [
    'load', 'save',
    'load_json', 'save_json',
    'load_pickle', 'save_pickle',
    'load_numpy', 'save_numpy',
    'load_csv', 'save_csv',
    'load_text', 'save_text',
]


def load(source: Union[str, PathLike, BinaryIO], hint: MaybeHint = None, **kwargs) -> Any:
    """
    Load a value from a file-like or buffer `source`.
    `hint` is used to override the format detection.
    `kwargs` are format-specific keyword arguments.
    """
    choice = Choice(*REGISTRY)

    if isinstance(source, (str, PathLike)):
        if hint is None:
            hint = os.path.basename(os.fspath(source))

        try:
            return choice.load_path(source, hint, kwargs)
        except WrongSerializer:
            pass

        with open(source, 'rb') as buffer:
            return choice.load_buffer(buffer, hint, False, kwargs)

    if not isinstance(source, BinaryIO):
        raise TypeError('Need a binary buffer')

    return choice.load_buffer(source, hint, True, kwargs)


def save(value: Any, destination: Union[str, PathLike, BinaryIO], hint: MaybeHint = None, **kwargs) -> Hint:
    """
    Save `value` to a file-like or buffer `destination`.
    `hint` is used to override the format detection.
    `kwargs` are format-specific keyword arguments.
    """
    choice = Choice(*REGISTRY)

    if isinstance(destination, (str, PathLike)):
        if hint is None:
            hint = os.path.basename(os.fspath(destination))

        try:
            return choice.save_path(value, destination, hint, kwargs)
        except WrongSerializer:
            pass

        with open(destination, 'wb') as buffer:
            return choice.save_buffer(value, buffer, hint, kwargs)

    if not isinstance(destination, BinaryIO):
        raise TypeError('Need a binary buffer')

    return choice.save_buffer(value, destination, hint, kwargs)


def load_json(path: PathLike):
    """Load the contents of a json file."""
    with open(path, 'r') as f:
        return json.load(f)


def save_json(value, path: PathLike, *, indent: int = None):
    """Dump a json-serializable object to a json file."""
    with open(path, 'w') as f:
        json.dump(value, f, indent=indent)


def save_numpy(value, path: PathLike, *, allow_pickle: bool = True, fix_imports: bool = True, compression: int = None,
               timestamp: int = None):
    """A wrapper around ``np.save`` that matches the interface ``save(what, where)``."""
    import numpy as np

    if compression is not None:
        with GzipFile(path, 'wb', compresslevel=compression, mtime=timestamp) as file:
            return save_numpy(value, file, allow_pickle=allow_pickle, fix_imports=fix_imports)

    np.save(path, value, allow_pickle=allow_pickle, fix_imports=fix_imports)


def load_numpy(path: PathLike, *, allow_pickle: bool = False, fix_imports: bool = True, decompress: bool = False):
    """A wrapper around ``np.load``"""
    import numpy as np

    if decompress:
        with GzipFile(path, 'rb') as file:
            return load_numpy(file, allow_pickle=allow_pickle, fix_imports=fix_imports)

    return np.load(path, allow_pickle=allow_pickle, fix_imports=fix_imports)


def save_pickle(value, path: PathLike):
    """Pickle a ``value`` to ``path``."""
    with open(path, 'wb') as file:
        pickle.dump(value, file)


def load_pickle(path: PathLike):
    """Load a pickled value from ``path``."""
    with open(path, 'rb') as file:
        return pickle.load(file)


def save_text(value: str, path: PathLike):
    with open(path, mode='w') as file:
        file.write(value)


def load_text(path: PathLike):
    with open(path, mode='r') as file:
        return file.read()


def save_csv(value, path: PathLike, *, compression: int = None, **kwargs):
    if compression is not None:
        kwargs['compression'] = {
            'method': 'gzip',
            'compresslevel': compression,
        }

    value.to_csv(path, **kwargs)


def load_csv(path: PathLike, **kwargs):
    import pandas as pd
    return pd.read_csv(path, **kwargs)



from abc import ABC, abstractmethod
from os import PathLike
from typing import Any, Union, BinaryIO, Tuple

from ..serializer import Serializer, MatchHint, MaybeHint, WrongSerializer, Hint


class NoBuffer(Serializer, ABC):
    def match_load_buffer(self, *args, **kwargs) -> MatchHint:
        return MatchHint.Reject

    def load_buffer(self, *args, **kwargs) -> MatchHint:
        raise WrongSerializer

    match_save_buffer = match_load_buffer
    save_buffer = load_buffer


class NoPath(Serializer, ABC):
    def match_load_path(self, *args, **kwargs) -> MatchHint:
        return MatchHint.Reject

    def load_path(self, *args, **kwargs) -> MatchHint:
        raise WrongSerializer

    match_save_path = match_load_path
    save_path = load_path


class SourceAgnostic(Serializer, ABC):
    @abstractmethod
    def load(self, source: Union[PathLike, BinaryIO], hint: MaybeHint, params) -> Any:
        pass

    def load_path(self, source: PathLike, hint: Hint, params: dict) -> Any:
        return self.load(source, hint, params)

    def load_buffer(self, source: BinaryIO, hint: MaybeHint, allow_lazy: bool, params: dict) -> Any:
        return self.load(source, hint, params)

    @abstractmethod
    def save(self, value: Any, destination: Union[PathLike, BinaryIO], hint: MaybeHint, params) -> Hint:
        pass

    def save_path(self, value: Any, destination: PathLike, hint: Hint, params: dict) -> Hint:
        return self.save(value, destination, hint, params)

    def save_buffer(self, value: Any, destination: BinaryIO, hint: MaybeHint, params: dict) -> Hint:
        return self.save(value, destination, hint, params)


class ExtensionMatch(Serializer, ABC):
    extensions: Tuple[str]

    def _match_name(self, name):
        return name is not None and any(name.endswith(x) for x in self.extensions)

    def _match_value(self, value):
        return True

    def _match_load_params(self, params: dict):
        return not params

    def _match_save_params(self, params: dict):
        return not params

    def match_save_buffer(self, value: Any, hint: MaybeHint, params: dict) -> MatchHint:
        return self._match_save(value, hint, params)

    def match_save_path(self, value: Any, destination: PathLike, hint: Hint, params: dict) -> MatchHint:
        return self._match_save(value, hint, params)

    def match_load_path(self, source: PathLike, hint: Hint, params: dict) -> MatchHint:
        return self._match_load(hint, params)

    def match_load_buffer(self, hint: MaybeHint, allow_lazy: bool, params: dict) -> MatchHint:
        return self._match_load(hint, params)

    def _match_save(self, value, hint, params):
        if not self._match_save_params(params):
            return MatchHint.Reject
        if not self._match_value(value):
            return MatchHint.Reject
        if not self._match_name(hint):
            return MatchHint.Reject
        return MatchHint.Accept

    def _match_load(self, hint, params):
        if not self._match_load_params(params):
            return MatchHint.Reject
        if not self._match_name(hint):
            return MatchHint.Reject
        return MatchHint.Accept



import json
import pickle
from gzip import GzipFile
from io import TextIOWrapper
from typing import Any, BinaryIO

try:
    from gzip import BadGzipFile
except ImportError:
    # for py3.6
    BadGzipFile = OSError

from .helpers import ExtensionMatch, NoPath
from ..serializer import Serializer, MatchHint, MaybeHint, WrongSerializer, REGISTRY, Hint


class JSON(NoPath, ExtensionMatch):
    extensions = '.json',

    def _match_save_params(self, params: dict):
        return set(params) <= {'indent'}

    def load_buffer(self, source: BinaryIO, hint: MaybeHint, allow_lazy: bool, params: dict) -> Any:
        wrapper = TextIOWrapper(source)
        try:
            return json.load(wrapper)
        except TypeError as e:
            raise WrongSerializer from e
        finally:
            wrapper.detach()

    def save_buffer(self, value: Any, destination: BinaryIO, hint: MaybeHint, params: dict) -> Hint:
        wrapper = TextIOWrapper(destination)
        try:
            json.dump(value, wrapper, **params)
        except TypeError as e:
            raise WrongSerializer from e
        finally:
            wrapper.detach()

        return '.json'


class Pickle(ExtensionMatch, NoPath):
    extensions = '.pkl',

    # TODO: check magic bits
    def load_buffer(self, source: BinaryIO, hint: MaybeHint, allow_lazy: bool, params: dict) -> Any:
        return pickle.load(source)

    def save_buffer(self, value: Any, destination: BinaryIO, hint: MaybeHint, params: dict) -> Hint:
        pickle.dump(value, destination)
        return '.pkl'


class Text(ExtensionMatch, NoPath):
    extensions = '.txt',

    def _match_value(self, value):
        return isinstance(value, str)

    def load_buffer(self, source: BinaryIO, hint: MaybeHint, allow_lazy: bool, params: dict) -> Any:
        wrapper = TextIOWrapper(source)
        try:
            return wrapper.read()
        finally:
            wrapper.detach()

    def save_buffer(self, value: Any, destination: BinaryIO, hint: MaybeHint, params: dict) -> Hint:
        wrapper = TextIOWrapper(destination)
        try:
            wrapper.write(value)
        finally:
            wrapper.detach()

        return '.txt'


class Gzip(NoPath, Serializer):
    def __init__(self, serializer: Serializer):
        self.serializer = serializer
        self._ext = '.gz'

    def _match(self, name):
        return name is not None and name.endswith(self._ext)

    def _trim(self, name):
        return None if name is None else name[:-len(self._ext)]

    def match_save_buffer(self, value: Any, hint: MaybeHint, params: dict) -> MatchHint:
        if not self._match(hint):
            return MatchHint.NotSure

        params = params.copy()
        params.pop('compression', None)
        return self.serializer.match_save_buffer(value, self._trim(hint), params)

    def match_load_buffer(self, hint: MaybeHint, allow_lazy: bool, params: dict) -> MatchHint:
        if not self._match(hint):
            return MatchHint.NotSure

        return self.serializer.match_load_buffer(self._trim(hint), allow_lazy, params)

    def load_buffer(self, source: BinaryIO, hint: MaybeHint, allow_lazy: bool, params: dict) -> Any:
        try:
            with GzipFile(fileobj=source, mode='rb') as local:
                return self.serializer.load_buffer(local, self._trim(hint), allow_lazy, params)
        except BadGzipFile as e:
            raise WrongSerializer from e

    def save_buffer(self, value: Any, destination: BinaryIO, hint: MaybeHint, params: dict) -> Hint:
        params = params.copy()
        compression = params.pop('compression', 1)
        mtime = 0

        try:
            with GzipFile(fileobj=destination, mode='wb', compresslevel=compression, mtime=mtime) as local:
                result = self.serializer.save_buffer(value, local, self._trim(hint), params)
        except BadGzipFile as e:
            raise WrongSerializer from e

        return result + self._ext


REGISTRY.extend((
    JSON(), Text(), Pickle(),
))



from os import PathLike
from typing import Any, BinaryIO, Sequence

from .. import Hint, MatchHint, MaybeHint
from ..serializer import Serializer, WrongSerializer, RequireLazy


class Choice(Serializer):
    def __init__(self, *choices: Serializer):
        self.choices = choices

    # matching

    def match_save_buffer(self, value: Any, hint: MaybeHint, params: dict) -> MatchHint:
        return min(
            (choice.match_save_buffer(value, hint, params) for choice in self.choices),
            key=lambda m: m.value, default=MatchHint.Reject
        )

    def match_save_path(self, value: Any, destination: PathLike, hint: Hint, params: dict) -> MatchHint:
        return min(
            (choice.match_save_path(value, destination, hint, params) for choice in self.choices),
            key=lambda m: m.value, default=MatchHint.Reject
        )

    def match_load_buffer(self, hint: MaybeHint, allow_lazy: bool, params: dict) -> MatchHint:
        return min(
            (choice.match_load_buffer(hint, allow_lazy, params) for choice in self.choices),
            key=lambda m: m.value, default=MatchHint.Reject
        )

    def match_load_path(self, source: PathLike, hint: Hint, params: dict) -> MatchHint:
        return min(
            (choice.match_load_path(source, hint, params) for choice in self.choices),
            key=lambda m: m.value, default=MatchHint.Reject
        )

    # saving

    def save_buffer(self, value: Any, destination: BinaryIO, hint: MaybeHint, params: dict) -> Hint:
        choices = self._filter(
            (choice.match_save_buffer(value, hint, params), choice)
            for choice in self.choices
        )

        position = destination.tell()
        for choice in choices:
            try:
                return choice.save_buffer(value, destination, hint, params)
            except WrongSerializer:
                if position != destination.tell():
                    destination.seek(position)
                    destination.truncate()

        raise WrongSerializer('No serializer was able to save the value')

    def save_path(self, value: Any, destination: PathLike, hint: Hint, params: dict) -> Hint:
        choices = self._filter(
            (choice.match_save_path(value, destination, hint, params), choice)
            for choice in self.choices
        )

        for choice in choices:
            try:
                return choice.save_path(value, destination, hint, params)
            except WrongSerializer:
                pass

        raise WrongSerializer('No serializer was able to save the value')

    # loading

    def load_buffer(self, source: BinaryIO, hint: MaybeHint, allow_lazy: bool, params: dict) -> Any:
        choices = self._filter(
            (choice.match_load_buffer(hint, allow_lazy, params), choice)
            for choice in self.choices
        )

        requires_lazy = None
        position = source.tell()
        for choice in choices:
            try:
                return choice.load_buffer(source, hint, allow_lazy, params)
            except (WrongSerializer, RequireLazy) as e:
                if isinstance(e, RequireLazy):
                    requires_lazy = e

                if position != source.tell():
                    source.seek(position)
                    source.truncate()

        if requires_lazy is not None:
            raise requires_lazy
        raise WrongSerializer('No serializer was able to save the value')

    def load_path(self, source: PathLike, hint: Hint, params: dict) -> Any:
        choices = self._filter(
            (choice.match_load_path(source, hint, params), choice)
            for choice in self.choices
        )

        for choice in choices:
            try:
                return choice.load_path(source, hint, params)
            except WrongSerializer:
                pass

        raise WrongSerializer('No serializer was able to save the value')

    # internals

    @staticmethod
    def _filter(values) -> Sequence[Serializer]:
        return [v for k, v in sorted(values, key=lambda m: m[0].value) if k is not MatchHint.Reject]



import runpy
from pathlib import Path

from setuptools import setup, find_packages  # type: ignore

NAME = 'deli'
DESCRIPTION = 'Smart serialization and deserialization for (almost) any python object'
LICENSE = 'MIT'
URL = 'https://github.com/maxme1/deli'

CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3 :: Only',
    'Operating System :: OS Independent',
]
KEYWORDS = ['serialization']

ROOT = Path(__file__).resolve().parent

VERSION = runpy.run_path(str(ROOT / NAME / '__version__.py'))['__version__']

with open(ROOT / 'README.md', 'r', encoding='utf-8') as file:
    LONG_DESCRIPTION = file.read()

with open(ROOT / 'requirements.txt', 'r', encoding='utf-8') as file:
    REQUIREMENTS = file.read().splitlines()

DOWNLOAD_URL = f'{URL}/archive/v{VERSION}.tar.gz'

setup(
    name=NAME,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    license=LICENSE,
    version=VERSION,
    url=URL,
    download_url=DOWNLOAD_URL,
    packages=find_packages(include=(NAME,)),
    include_package_data=True,
    install_requires=REQUIREMENTS,
    classifiers=CLASSIFIERS,
    keywords=KEYWORDS,
    python_requires='>=3.6',
)



from os import PathLike
from typing import Any

from .helpers import ExtensionMatch, NoBuffer
from .packaged import Gzip
from ..serializer import REGISTRY, Hint


class Nifty(ExtensionMatch, NoBuffer):
    extensions = '.nii', '.nii.gz'

    def _match_value(self, value):
        return isinstance(value, Nifti1Image)

    def load_path(self, source: PathLike, hint: Hint, params: dict) -> Any:
        return nibabel.load(source, **params)

    def save_path(self, value: Any, destination: PathLike, hint: Hint, params: dict) -> Hint:
        nibabel.save(value, destination, **params)
        if not hint.endswith(self.extensions):
            return '.nii'
        return hint


# TODO:
# load from buffer
# nii = nibabel.FileHolder(fileobj=buffer)
# return Nifti1Image.from_file_map({'header': nii, 'image': nii})

# save to buffer
# buffer.write(value.to_bytes())
# with  Opener(BytesIO()) as bio:
#     # file_map = value.make_file_map({'image': bio, 'header': bio})
#     file_map = {'image': nibabel.FileHolder(fileobj=bio), 'header': nibabel.FileHolder(fileobj=bio)}
#     print(file_map)
#     value.to_file_map(file_map)
#     buffer.write(bio.getvalue())


try:
    import nibabel
    from nibabel import Nifti1Image

    REGISTRY.append(Nifty())
    REGISTRY.append(Gzip(Nifty()))

except ImportError:
    pass
